#!/usr/bin/python

'''
   File name: TlogLib.py
   Author: Wayne Parham
   Date created: 09/01/2019
'''


import json, re, string, struct, sys


def ParseTlog( tlog ):
   tloglist = []
   fieldlist = []
   recordlist = []
   for line in tlog:
      line = CleanLine( line )
      records = ParseLine( line )
      for record in records:
         f = 0
         field_types = []
         metadata_available = False
         stringtype, subtype = Identify( record )
         field_types = IdentifyFields( stringtype, subtype )
         if( sys.version_info[0] < 3 ):
            fields = string.split(record, ':')
         else:
            fields = record.split( b':' )
         if( len(field_types) > 0 ):
            metadata_available = True
         for field in fields:
            if( metadata_available ):
               try:
                  field_desc = field_types[f][0]
                  field_type = field_types[f][1][0].lower()
                  field_len  = int( field_types[f][1][1:] )
               except:
                  field_desc = "field " + str(f)
                  field_type = "p"
                  field_len = len( field )
            else:
               field_desc = "field " + str(f)
               field_type = "p"
               field_len = len( field )
            # Append field_desc for JSON key/value pairs
            fieldlist.append( field )
            f += 1
         recordlist.append( fieldlist )
         fieldlist = []
      tloglist.append( recordlist )
      recordlist = []
   return( tloglist )


def Tlog2Csv( tloglist ):
   count = 0
   csvtlog = ""
   fieldlist = []
   recordlist = []
   for line in tloglist:
      count += 1
      csvtlog += "### Record # " + str(count) + " ###\n"
      for record in line:
         for field in record:
            if( sys.version_info[0] < 3 ):
               field = re.sub( r'bytearray\((.*)\)', r'\1', repr(field) )
               fieldlist.append( field )
            else:
               fieldlist.append( str(field) )
         csvtlog += CSVify( fieldlist )
         fieldlist = []
   return( csvtlog )


def Tlog2Json( tloglist ):
   jsonlist = []
   fieldlist = []
   recordlist = []
   for line in tloglist:
      for record in line:
         for field in record:
            try:
               decoded_field = Strify( field )
            except:
               decoded_field = field
            fieldlist.append( decoded_field )
         recordlist.append( fieldlist )
         fieldlist = []
      jsonlist.append( recordlist )
      recordlist = []
   jsontlog = json.dumps( jsonlist, indent = 3 )
   return( jsontlog )


def GetDataFromFile( filename ):
   tlogfile = open(filename, "rb")
   data = tlogfile.readlines()
   tlogfile.close()
   return( data )


def GetDataFromStdin():
   if( sys.version_info[0] < 3 ):
      data = sys.stdin.readlines()
   else:
      data = sys.stdin.buffer.readlines()
   return( data )


def CleanLine( line ):
   line = line.rstrip()
   if( sys.version_info[0] < 3 ):
      if( line[-1] == '"' ):
         line = line[:-1]
      if( line[0] == '"' ):
         line = line[1:]
   else:
      if( line[-1] == ord('"') ):
         line = line[:-1]
      if( line[0] == ord('"') ):
         line = line[1:]
   return( line )


def ParseLine( line ):
   if( sys.version_info[0] < 3 ):
      records = []
      srecords = line.split( '","' )
      for srecord in srecords:
         record = bytearray( srecord )
         records.append( record )
   else:
      records = line.split( b'","' )
   return( records )


def Identify( record ):
   try:
      stringtype = hex(ord(record[0]))
   except:
      stringtype = None
   try:
      subtype = hex(ord(record[1]))
   except:
      subtype = None
   return( stringtype, subtype )


def IdentifyFields( stringtype, subtype=None ):
   # get definitions from gsa.ini or tlogstrings.xml
   field_types = []
   if( stringtype == "0x0" ):
      # Header String
      field_types = [
                      ["String Type", "p1"],
                      ["Terminal Number", "p2"],
                      ["Operator Number", "p4"],
                      ["Transaction Number", "p2"],
                      ["Transaction Type", "p1"],
                      ["Number of Strings in Transaction", "p2"],
                      ["Gross Positive", "p5"],
                      ["Gross Negative", "p5"],
                      ["Net Cash", "p4"],
                      ["Net Non-Cash", "p4"],
                      ["Password", "p3"],
                      ["Date", "p3"],
                      ["Time", "p2"],
                      ["Indicator 0", "p3"],
                      ["Indicator 1", "p3"],
                      ["Ring Time (secs)", "p4"],
                      ["Tender Time (secs)", "p4"],
                      ["Other Time (secs)", "p4"],
                      ["Prev Tend Time (secs)", "p4"],
                      ["Indicator 2", "p3"],
                      ["Number of Tenant Customer", "p3"],
                      ["POS Credit Apply Time (secs)", "p4"],
                      ["Terminal Inactive Time (secs)", "p4"],
                      ["Transaction Token", "a44"],
                      ["Token Type", "p1"]
                    ]
   elif( stringtype == "0x6" ):
      # Item String
      field_types = [
                      ["String Type", "p1"],
                      ["UPC Number", "p11"],
                      ["Department Number", "p3"],
                      ["Extended Amount", "p4"],
                      ["Tax Flags", "p1"],
                      ["Indicator 0", "p3"],
                      ["Indicator 1", "p2"],
                      ["Report Code", "p1"],
                      ["Indicator 1.2", "p4"],
                      ["FBL Indicat0", "p1"],
                      ["FBL Indicat1", "p1"],
                      ["FBL Indicat2", "p1"],
                      ["Item Description", "a12"],
                      ["FBL User Area", "p8"],
                      ["Wal-Mart Item Number", "p5"],
                      ["RX Number", "p6"],
                      ["Indicator 2", "p4"],
                      ["Indicator 3", "p4"],
                      ["Refund Group", "p1"],
                      ["Extended Barcode", "p3"],
                      ["Barcode Type", "p3"],
                      ["Link Save Offer 1", "p2"],
                      ["Link Save Offer 2", "p2"],
                      ["Restriction", "p2"],
                      ["Merchandise Category", "p2"],
                      ["Merchandise Subcategory", "p2"],
                      ["Extended Cost", "p4"],
                      ["WIC Category", "p2"],
                      ["WIC Subcategory", "p2"],
                      ["WIC Units Debited", "p4"],
                      ["WIC Subcategory Secondary", "p2"],
                      ["WIC Units Secondary", "p4"],
                      ["WIC Subcategory Tertiary", "p2"],
                      ["WIC Units Tertiary", "p4"],
                      ["Promo ID", "p2"],
                      ["EPC ID", "a28"],
                      ["Long Item Description", "a25"]
                    ]
   elif( stringtype == "0x15" ):
      if( subtype == "0x13" ):
         # Product Serial Number
         field_types = [
                      ["String Type", "p1"],
                      ["Substring Type", "p1"],
                      ["UPC Number", "a14"],
                      ["Product Serial Number", "a30"],
                      ["Sale (0) or Void (1)", "a1"],
                      ["Company Code", "a2"],
                      ["Item Extended Amount", "p5"],
                      ["Authorization Number", "p5"],
                      ["Indicat0", "p1"],
                      ["Insured Item UPC", "a14"],
                      ["Insured Item Purchase Date", "p4"],
                      ["Starting Date of Warranty", "p4"],
                      ["Ending Date of Warranty", "p4"],
                      ["Manufacturer Warranty Period", "p1"],
                      ["Extended Warranty Period", "p1"]
                       ]
   return( field_types )


def Asciify( record ):
   asciir = []
   stringtype, subtype = Identify( record )
   field_types = IdentifyFields( stringtype, subtype )
   if( len(field_types) > 0 ):
      """
      f = 0
      for field_type in field_types:
         field_desc = field_types[f][0]
         field_type = field_types[f][1][0].lower()
         field_len  = int( field_types[f][1][1:] )
         f += 1
      """
      f = 0
      for field in record:
         try:
            if( field_types[f][1][0].lower() == "p" ):
               asciir.append( BCD2string( field ) )
            else:
               asciir.append( field )
         except:
            asciir.append( field )
      else:
         asciir.append( field )
   return( asciir )


def Bytify( string ):
   bytelist = []
   for c in string:
      n = ord(c)
      if( n <= 0xFF ):
         bytelist.append(struct.pack('B', n))
      else:
         bytelist.append(struct.pack('>H', n))
   return b''.join(bytelist)


def Strify( bytestring ):
   string = ""
   try:
      string = bytestring.decode( 'latin1' )
   except:
      string = bytestring
   return( string )


def CSVify( items ):
   c = 0
   csvline = ""
   last = len( items )
   for item in items:
      c += 1
      csvline += item
      if( c < last ):
         csvline += ","
      else:
         csvline += "\n"
   return( csvline )


def BCD2string( bytestring ):
   # leftmost "F" means positive
   # leftmost "D" means negative
   string = ""
   for byte in bytestring:
      sign = None
      left = int( byte/16 )
      right = byte % 16
      if not sign:
         if( left == 0xD ):
            sign = "-"
         else:
            sign= "+"
      if( left > 9):
         left = 0
      if( right > 9 ):
          right = 0
      string += str(left)
      string += str(right)
   if( sign == "-" ):
      string = sign + string
   return( string )

